#include "params.h"

using namespace MaskTools::Clip;
using namespace MaskTools::Parameters;
using namespace MaskTools::Constraints;

namespace MaskTools { namespace Avs25 {
   
Clip::Clip AVSValueToClip(const AVSValue &value)
{
   if ( value.IsClip() )
   {
      int nFrames = value.AsClip()->GetVideoInfo().num_frames;
      int nWidth = value.AsClip()->GetVideoInfo().width;
      int nHeight = value.AsClip()->GetVideoInfo().height;
      return Clip::Clip(nWidth, nHeight, nFrames);
   }
   else
      return Clip::Clip();
}

Operator AVSValueToOperator(const AVSValue &value, const Operator &def_op)
{
   if ( value.Defined() && value.IsInt() )
   {
      switch ( value.AsInt() )
      {
      case 4: return Operator(COPY_SECOND);
      case 3: return Operator(PROCESS);
      case 2: return Operator(COPY);
      case 1: return Operator(NONE);
      default: return Operator(-value.AsInt());
      }
   }
   else
      return def_op;
}

Value AVSValueToValue(const AVSValue& value, const Parameter &param)
{
   switch ( Type(Value(param)) )
   {
   case Parameters::INT: return Value(value.AsInt());
   case Parameters::STRING: return Value(String(value.AsString()));
   case Parameters::BOOL: return Value(value.AsBool());
   case Parameters::OPERATOR: return AVSValueToOperator(value, Value(param));
   case Parameters::CLIP: return AVSValueToClip(value);
   case Parameters::FLOAT: return Value(value.AsFloat());
   }
   return Value();
}

String ParamToString(const Parameter& parameter)
{
   String str = "";

   if ( String(parameter).length() )
      str.append("[").append(parameter).append("]");

   switch ( Type(parameter) )
   {
   case Parameters::INT      : return str.append("i");
   case Parameters::OPERATOR : return str.append("i");
   case Parameters::FLOAT    : return str.append("f");
   case Parameters::STRING   : return str.append("s");
   case Parameters::CLIP     : return str.append("c");
   case Parameters::BOOL     : return str.append("b");
   }

   print(LOG_ERROR, "unvalid parameter type : %i", Type(parameter));

   return "";
}

String SignatureToParamsSignature(const Signature &signature)
{
   String str_params = "";

   for ( Signature::const_iterator it = signature.begin(); it != signature.end(); it++ )
      str_params.append(ParamToString(*it));

   return str_params;
}

static Parameters::Parameter &GetParameter(Parameters::Parameter &param, const AVSValue &value, const Parameters::Parameter &default_param)
{
   if ( value.Defined() )
      param = Parameters::Parameter(AVSValueToValue(value, default_param), String(default_param));
   else
   {
      param = default_param;
      param.define(false);
   }

   return param;
}

Parameters::Parameters GetParameters(const AVSValue &args, const Signature &signature, IScriptEnvironment *env)
{
   const int nCount = int(signature.size()) - 13; /* 11 common parameters */
   Parameters::Parameters params(nCount);
   Parameters::Parameter param;

#define MT_GET_VALUE(n) GetParameter(params[n], args[n], signature[n])

   /* specific parameters */
   for ( int nIdx = 0; nIdx < nCount; nIdx ++ )
      MT_GET_VALUE(nIdx);

#undef MT_GET_VALUE
#define MT_GET_VALUE(n) GetParameter(param, args[n + nCount], signature[n + nCount]) 

   /* common settings */
   params.Y = MT_GET_VALUE(0);
   params.U = MT_GET_VALUE(1);
   params.V = MT_GET_VALUE(2);

   params.nXOffset = MT_GET_VALUE(3);
   params.nYOffset = MT_GET_VALUE(4);
   params.nWidth   = MT_GET_VALUE(5);
   params.nHeight  = MT_GET_VALUE(6);

   bool mmx     = Value(MT_GET_VALUE(7)).toBool();
   bool isse    = Value(MT_GET_VALUE(8)).toBool();
   bool sse2    = Value(MT_GET_VALUE(9)).toBool();
   bool sse3    = Value(MT_GET_VALUE(10)).toBool();          /* to change, if avisynth eevolves to detect sse3 */
   bool _3dnow  = Value(MT_GET_VALUE(11)).toBool();
   bool _3dnow2 = Value(MT_GET_VALUE(12)).toBool();

#undef MT_GET_VALUE

   /* check available cpu flags */
   mmx      = mmx    && (env->GetCPUFlags() & CPUF_MMX);
   isse     = isse   && (env->GetCPUFlags() & CPUF_INTEGER_SSE);
   sse2     = sse2   && (env->GetCPUFlags() & CPUF_SSE2);
   _3dnow   = _3dnow && (env->GetCPUFlags() & CPUF_3DNOW);
   _3dnow2  = _3dnow2&& (env->GetCPUFlags() & CPUF_3DNOW_EXT);

   CpuFlags flags = CPU_NONE;
   flags = CpuFlags(flags | (    mmx ? CPU_MMX    : CPU_NONE));
   flags = CpuFlags(flags | (   isse ? CPU_ISSE   : CPU_NONE));
   flags = CpuFlags(flags | (   sse2 ? CPU_SSE2   : CPU_NONE));
   flags = CpuFlags(flags | (   sse3 ? CPU_SSE3   : CPU_NONE));
   flags = CpuFlags(flags | ( _3dnow ? CPU_3DNOW  : CPU_NONE));
   flags = CpuFlags(flags | (_3dnow2 ? CPU_3DNOW2 : CPU_NONE));

   params.flags = int(flags);

   /* clip settings */

   /* assume first parameter to be a clip */
   if ( !args[0].IsClip() )
      print(LOG_ERROR, "First parameter must be a clip\n");

   params.clip = AVSValueToClip(args[0]);

   return params;
}



}
}