#include "symbol.h"
#include <math.h>

using namespace MaskTools;
using namespace MaskTools::Parser;

static double addition        (double x, double y, double z) { return x + y; }
static double multiplication  (double x, double y, double z) { return x * y; }
static double division        (double x, double y, double z) { return x / y; }
static double substraction    (double x, double y, double z) { return x - y; }
static double power           (double x, double y, double z) { return pow(x, y); }
static double modulo          (double x, double y, double z) { return double(int(x + 0.5) % int(y + 0.5)); }
static double interrogation   (double x, double y, double z) { return x > 0 ? y : z; }
static double equal           (double x, double y, double z) { return x == y ? 1 : -1; }
static double notEqual        (double x, double y, double z) { return x != y ? 1 : -1; }
static double inferior        (double x, double y, double z) { return x <= y ? 1 : -1; }
static double inferiorStrict  (double x, double y, double z) { return x < y ? 1 : -1; }
static double superior        (double x, double y, double z) { return x >= y ? 1 : -1; }
static double superiorStrict  (double x, double y, double z) { return x > y ? 1 : -1; }
static double and             (double x, double y, double z) { return x > 0 && y > 0 ? 1 : -1; }
static double or              (double x, double y, double z) { return x > 0 || y > 0 ? 1 : -1; }
static double andNot          (double x, double y, double z) { return x > 0 && y <= 0 ? 1 : -1; }
static double xor             (double x, double y, double z) { return (x > 0 && y <= 0) || (x <= 0 && y > 0)? 1 : -1; }
static double cos             (double x, double y, double z) { return cos(x); }
static double sin             (double x, double y, double z) { return sin(x); }
static double tan             (double x, double y, double z) { return tan(x); }
static double exp             (double x, double y, double z) { return exp(x); }
static double log             (double x, double y, double z) { return log(x); }
static double abs             (double x, double y, double z) { return abs(x); }
static double acos            (double x, double y, double z) { return acos(x); }
static double asin            (double x, double y, double z) { return asin(x); }
static double atan            (double x, double y, double z) { return atan(x); }

Symbol Symbol::Addition       ("+" , OPERATOR, 2, addition);
Symbol Symbol::Multiplication ("*" , OPERATOR, 2, multiplication);
Symbol Symbol::Division       ("/" , OPERATOR, 2, division);
Symbol Symbol::Substraction   ("-" , OPERATOR, 2, substraction);
Symbol Symbol::Power          ("^" , OPERATOR, 2, power);
Symbol Symbol::Modulo         ("%" , OPERATOR, 2, modulo);
Symbol Symbol::Interrogation  ("?" , TERNARY , 3, interrogation);
Symbol Symbol::Equal          ("==", OPERATOR, 2, equal);
Symbol Symbol::NotEqual       ("!=", OPERATOR, 2, notEqual);
Symbol Symbol::Inferior       ("<=", OPERATOR, 2, inferior);
Symbol Symbol::InferiorStrict ("<" , OPERATOR, 2, inferiorStrict);
Symbol Symbol::Superior       (">=", OPERATOR, 2, superior);
Symbol Symbol::SuperiorStrict (">" , OPERATOR, 2, superiorStrict);
Symbol Symbol::And            ("&" , OPERATOR, 2, and);
Symbol Symbol::Or             ("|" , OPERATOR, 2, or);
Symbol Symbol::AndNot         ("&!", OPERATOR, 2, andNot);
Symbol Symbol::Xor            ("" , OPERATOR, 2, xor);
Symbol Symbol::Pi             ("pi", 3.1415927, NUMBER  , 0, NULL);
Symbol Symbol::X              ("x" , VARIABLE, 0, NULL);  
Symbol Symbol::Y              ("y" , VARIABLE, 0, NULL);
Symbol Symbol::Cos            ("cos", FUNCTION, 1, cos);
Symbol Symbol::Sin            ("sin", FUNCTION, 1, sin);
Symbol Symbol::Tan            ("tan", FUNCTION, 1, tan);
Symbol Symbol::Log            ("log", FUNCTION, 1, log);
Symbol Symbol::Exp            ("exp", FUNCTION, 1, exp);
Symbol Symbol::Abs            ("abs", FUNCTION, 1, abs);
Symbol Symbol::Atan           ("atan", FUNCTION, 1, atan);
Symbol Symbol::Acos           ("acos", FUNCTION, 1, acos);
Symbol Symbol::Asin           ("asin", FUNCTION, 1, asin);

Symbol::Symbol() :
type(UNDEFINED), value("")
{
}

Symbol::Symbol(String value, Type type, int nParameter, Process process) :
type(type), value(value), nParameter(nParameter), process(process)
{
   if ( type == NUMBER )
      dValue = atof(value.c_str());
}

Symbol::Symbol(String value, double dValue, Type type, int nParameter, Process process) :
type(type), value(value), nParameter(nParameter), process(process), dValue(dValue)
{
}

void Symbol::setValue(double dValue)
{
   this->dValue = dValue;
}

double Symbol::getValue(double x, double y, double z) const
{
   switch ( type )
   {
   case OPERATOR:
   case FUNCTION:
   case TERNARY:
      return process(x, y, z);
   case VARIABLE:
   case NUMBER:
      return dValue;
   }

   return -1.0f;
}

Context::Context(const std::list<Symbol> &expression) :
expression(expression)
{
}

Context &Context::setVariable(const String &name, double value)
{
   for ( std::list<Symbol>::iterator it = expression.begin(); it != expression.end(); it++ )
      if ( it->type == Symbol::VARIABLE && it->value == name )
         it->setValue(value);

   return *this;
}

double Context::popb()
{
   if ( size() )
   {
      double f = back();
      pop_back();
      return f;
   }
   return -1.0f;
}

double Context::compute()
{
   std::list<Symbol> exp = expression;

   clear();

   while ( exp.size() )
   {
      Symbol s = exp.front();
      exp.pop_front();

      if ( s.nParameter > int(size()) )
      {
         print(LOG_ERROR, "unvalid expression : not enough values stacked");
         return -1.0f;
      }

      double z = s.nParameter >= 3 ? popb() : -1.0f;
      double y = s.nParameter >= 2 ? popb() : -1.0f;
      double x = s.nParameter >= 1 ? popb() : -1.0f;

      push_back(s.getValue(x, y, z));
   }

   if ( size() > 1 )
   {
      print(LOG_ERROR, "still some value on the stack");
      return -1.0f;
   }

   if ( !size() )
   {
      print(LOG_ERROR, "no values left on the stack");
      return -1.0f;
   }

   return front();
}

bool Context::check()
{
   std::list<Symbol> exp = expression;

   clear();

   while ( exp.size() )
   {
      Symbol s = exp.front();
      exp.pop_front();

      if ( s.nParameter > int(size()) )
         return false;

      double z = s.nParameter >= 3 ? popb() : -1.0f;
      double y = s.nParameter >= 2 ? popb() : -1.0f;
      double x = s.nParameter >= 1 ? popb() : -1.0f;

      push_back(s.getValue(x, y, z));
   }

   if ( size() > 1 )
      return false;

   if ( !size() )
      return false;

   return true;
}